@@ -2,26 +2,25 @@
 # SPDX-FileCopyrightText: 2022 geisserml <geisserml@gmail.com>
 # SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause
 
-# Attempt to download and build PDFium from source. This may take very long.
-# Only tested on Linux with glibc. Might not work on other platforms.
-
 import os
 import sys
 import shutil
 import argparse
-import urllib.request
 from os.path import join, abspath, dirname
 
 sys.path.insert(0, dirname(dirname(abspath(__file__))))
 from pl_setup.packaging_base import (
+    Host,
     SB_Dir,
-    Libnames,
     DataTree,
-    VerNamespace,
+    PDFium_URL,
+    DepotTools_URL,
+    MainLibnames,
+    LibnameForSystem,
+    VerStatusFileName,
     PlatformNames,
     run_cmd,
     call_ctypesgen,
-    set_version,
 )
 
 
@@ -31,16 +30,13 @@ PDFiumDir      = join(SB_Dir, "pdfium")
 PDFiumBuildDir = join(PDFiumDir, "out", "Default")
 OutputDir      = join(DataTree, PlatformNames.sourcebuild)
 
-DepotTools_URL = "https://chromium.googlesource.com/chromium/tools/depot_tools.git"
-PDFium_URL     = "https://pdfium.googlesource.com/pdfium.git"
-
 PdfiumMainPatches = [
-    (join(PatchDir, "pdfium", "public_headers.patch"), PDFiumDir),
-    (join(PatchDir, "pdfium", "shared_library.patch"), PDFiumDir),
+    (join(PatchDir, "public_headers.patch"), PDFiumDir),
+    (join(PatchDir, "shared_library.patch"), PDFiumDir),
 ]
 PdfiumWinPatches = [
-    (join(PatchDir, "pdfium", "win", "pdfium.patch"), PDFiumDir),
-    (join(PatchDir, "pdfium", "win", "build.patch"), join(PDFiumDir, "build")),
+    (join(PatchDir, "win", "pdfium.patch"), PDFiumDir),
+    (join(PatchDir, "win", "build.patch"), join(PDFiumDir, "build")),
 ]
 
 DefaultConfig = {
@@ -69,21 +65,21 @@ def dl_depottools(do_update):
     if os.path.isdir(DepotToolsDir):
         if do_update:
             print("DepotTools: Revert and update ...")
-            run_cmd([Git, "reset", "--hard", "HEAD"], cwd=DepotToolsDir)
-            run_cmd([Git, "pull", DepotTools_URL], cwd=DepotToolsDir)
+            run_cmd(["git", "reset", "--hard", "HEAD"], cwd=DepotToolsDir)
+            run_cmd(["git", "pull", DepotTools_URL], cwd=DepotToolsDir)
         else:
             print("DepotTools: Using existing repository as-is.")
             is_update = False
     else:
         print("DepotTools: Download ...")
-        run_cmd([Git, "clone", "--depth", "1", DepotTools_URL, DepotToolsDir], cwd=SB_Dir)
+        run_cmd(["git", "clone", "--depth", "1", DepotTools_URL, DepotToolsDir], cwd=SB_Dir)
     
     os.environ["PATH"] += os.pathsep + DepotToolsDir
     
     return is_update
 
 
-def dl_pdfium(do_update, revision):
+def dl_pdfium(GClient, do_update, revision):
     
     is_sync = True
     
@@ -105,8 +101,8 @@ def dl_pdfium(do_update, revision):
 
 
 def get_version_info():
-    head_commit = run_cmd([Git, "rev-parse", "--short", "HEAD"], cwd=PDFiumDir, capture=True)
-    refs_string = run_cmd([Git, "ls-remote", "--heads", PDFium_URL, "chromium/*"], cwd=None, capture=True)
+    head_commit = run_cmd(["git", "rev-parse", "--short", "HEAD"], cwd=PDFiumDir, capture=True)
+    refs_string = run_cmd(["git", "ls-remote", "--heads", PDFium_URL, "chromium/*"], cwd=None, capture=True)
     latest = refs_string.split("\n")[-1]
     tag_commit, ref = latest.split("\t")
     tag = ref.split("/")[-1]
@@ -115,25 +111,20 @@ def get_version_info():
 
 def update_version(head_commit, tag_commit, tag):
     
-    is_sourcebuild  = VerNamespace["IS_SOURCEBUILD"]
-    curr_libversion = VerNamespace["V_LIBPDFIUM"]
-    
     print("Current head %s, latest tagged commit %s (%s)" % (head_commit, tag_commit, tag))
     if head_commit == tag_commit:
-        new_libversion = tag
+        v_libpdfium = tag
     else:
-        new_libversion = head_commit
+        v_libpdfium = head_commit
     
-    if not is_sourcebuild:
-        print("Switching from pre-built binaries to source build.")
-        set_version("IS_SOURCEBUILD", True)
-    if curr_libversion != new_libversion:
-        set_version("V_LIBPDFIUM", new_libversion)
+    ver_file = join(OutputDir, VerStatusFileName)
+    with open(ver_file, "w") as fh:
+        fh.write( str(v_libpdfium) )
 
 
 def _apply_patchset(patchset):
     for patch, cwd in patchset:
-        run_cmd([Git, "apply", "-v", patch], cwd=cwd)
+        run_cmd(["git", "apply", "-v", patch], cwd=cwd)
 
 def patch_pdfium():
     _apply_patchset(PdfiumMainPatches)
@@ -142,7 +133,7 @@ def patch_pdfium():
         shutil.copy(join(PatchDir, "pdfium", "win", "resources.rc"), join(PDFiumDir, "resources.rc"))
 
 
-def configure(config):
+def configure(GN, config):
     if not os.path.exists(PDFiumBuildDir):
         os.makedirs(PDFiumBuildDir)
     with open(join(PDFiumBuildDir, "args.gn"), "w") as args_handle:
@@ -150,58 +141,55 @@ def configure(config):
     run_cmd([GN, "gen", PDFiumBuildDir], cwd=PDFiumDir)
 
 
-def build(target):
+def build(Ninja, target):
     run_cmd([Ninja, "-C", PDFiumBuildDir, target], cwd=PDFiumDir)
 
 
-def find_lib(srcname=None, directory=PDFiumBuildDir):
+def find_lib(src_libname=None, directory=PDFiumBuildDir):
     
-    if srcname is not None:
-        path = join(PDFiumBuildDir, srcname)
+    if src_libname is not None:
+        path = join(PDFiumBuildDir, src_libname)
         if os.path.isfile(path):
             return path
         else:
-            print("Warning: The file of given srcname does not exist.", file=sys.stderr)
+            print("Warning: Binary not found under given name.", file=sys.stderr)
     
-    libpath = None
-    for lname in Libnames:
-        path = join(directory, lname)
-        if os.path.isfile(path):
-            libpath = path
+    try_names = []
+    for name in MainLibnames + ["pdfium.so"]:
+        try_names += [name, "lib"+name]
     
-    if libpath is None:
-        raise RuntimeError("Build artifact not found.")
+    try_paths = [join(directory, n) for n in try_names]
+    found_paths = [fp for fp in try_paths if os.path.isfile(fp)]
     
-    return libpath
+    assert len(found_paths) == 1
+    return found_paths[0]
 
 
 def pack(src_libpath, destname=None):
     
-    if os.path.isdir(OutputDir):
-        shutil.rmtree(OutputDir)
-    os.makedirs(OutputDir)
+    # TODO remove existing binary/bindings, just to be safe
     
     if destname is None:
-        destname = os.path.basename(src_libpath)
+        destname = LibnameForSystem[Host.system]
     
     destpath = join(OutputDir, destname)
     shutil.copy(src_libpath, destpath)
     
-    include_dir = join(OutputDir, "include")
-    shutil.copytree(join(PDFiumDir, "public"), include_dir)
+    ver_info = get_version_info()
+    update_version(*ver_info)
     
+    include_dir = join(PDFiumDir, "public")
     call_ctypesgen(OutputDir, include_dir)
-    shutil.rmtree(include_dir)
 
 
-def _get_tool(tool, win_append):
+def get_tool(tool, win_append):
     exe = join(DepotToolsDir, tool)
     if sys.platform.startswith("win32"):
         exe += "." + win_append
     return exe
 
 
-def _serialise_config(config_dict):
+def serialise_config(config_dict):
     
     config_str = ""
     sep = ""
@@ -220,48 +208,42 @@ def _serialise_config(config_dict):
 
 
 def main(
-        b_srcname = None,
-        b_destname = None,
+        b_src_libname = None,
+        b_dest_libname = None,
         b_update = False,
         b_revision = None,
         b_target = None,
     ):
     
-    global Git
-    Git = shutil.which("git")
+    if not os.path.exists(OutputDir):
+        os.makedirs(OutputDir)
     
     if b_revision is None:
         b_revision = "main"
     if b_target is None:
         b_target = "pdfium"
-    # on Linux, rename the binary to `pdfium` to ensure it also works with older versions of ctypesgen
-    if b_destname is None and sys.platform.startswith("linux"):
-        b_destname = "pdfium"
     
     if sys.platform.startswith("win32"):
         os.environ["DEPOT_TOOLS_WIN_TOOLCHAIN"] = "0"
     
     dl_depottools(b_update)
     
-    global GClient, GN, Ninja
-    GClient = _get_tool("gclient", "bat")
-    GN      = _get_tool("gn", "bat")
-    Ninja   = _get_tool("ninja", "exe")
+    GClient = get_tool("gclient", "bat")
+    GN      = get_tool("gn", "bat")
+    Ninja   = get_tool("ninja", "exe")
     
-    pdfium_dl_done = dl_pdfium(b_update, b_revision)
+    pdfium_dl_done = dl_pdfium(GClient, b_update, b_revision)
     if pdfium_dl_done:
         patch_pdfium()
     
-    update_version( *get_version_info() )
-    
     config_dict = DefaultConfig.copy()
-    config_str = _serialise_config(config_dict)
+    config_str = serialise_config(config_dict)
     print("\nBuild configuration:\n%s\n" % config_str)
     
-    configure(config_str)
-    build(b_target)
-    libpath = find_lib(b_srcname)
-    pack(libpath, b_destname)
+    configure(GN, config_str)
+    build(Ninja, b_target)
+    libpath = find_lib(b_src_libname)
+    pack(libpath, b_dest_libname)
 
 
 def parse_args(argv):
@@ -271,12 +253,12 @@ def parse_args(argv):
     )
     
     parser.add_argument(
-        "--srcname", "-s",
+        "--src-libname",
         help = "Name of the generated PDFium binary file. This script tries to automatically find the binary, which should usually work. If it does not, however, this option may be used to explicitly provide the file name to look for.",
     )
     parser.add_argument(
-        "--destname", "-d",
-        help = "Rename the binary to a different filename.",
+        "--dest-libname",
+        help = "Rename the binary. Must be a name recognised by packaging code.",
     )
     parser.add_argument(
         "--update", "-u",
@@ -298,11 +280,11 @@ def parse_args(argv):
 def main_cli(argv=sys.argv[1:]):
     args = parse_args(argv)
     return main(
-        b_srcname = args.srcname,
-        b_destname = args.destname,
-        b_update = args.update,
+        b_src_libname  = args.src_libname,
+        b_dest_libname = args.dest_libname,
+        b_update   = args.update,
         b_revision = args.revision,
-        b_target = args.target,
+        b_target   = args.target,
     )
     
 
