@@ -29,7 +29,7 @@ def attach_parser(subparsers):
 
 def main(args):
     
-    doc = pdfium.PdfDocument(args.input, args.password)
+    doc = pdfium.PdfDocument(args.input, password=args.password)
     if args.pages is None:
         args.pages = [i for i in range(len(doc))]
     
@@ -40,7 +40,7 @@ def main(args):
         textpage = page.get_textpage()
         text = textpage.get_text()
         
-        [g.close() for g in (textpage, page)]
+        for g in (textpage, page): g.close()
         
         print(sep + "# Page %s\n" % (index+1) + text)
         sep = "\n"
