@@ -3,10 +3,12 @@
 # SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause
 
 import sys
+import logging
 import argparse
-from pypdfium2._version import (
+from pypdfium2.version import (
     V_PYPDFIUM2,
     V_LIBPDFIUM,
+    IS_SOURCEBUILD,
 )
 from pypdfium2._cli import (
     render,
@@ -20,9 +22,7 @@ from pypdfium2._cli import (
 try:
     import argcomplete
 except ImportError:
-    have_argcomplete = False
-else:
-    have_argcomplete = True
+    argcomplete = None
 
 
 Subcommands = {
@@ -44,22 +44,30 @@ def parse_args(argv=sys.argv[1:]):
     parser.add_argument(
         "--version", "-v",
         action = "version",
-        version = "pypdfium2 %s (libpdfium %s)" % (V_PYPDFIUM2, V_LIBPDFIUM),
+        version = "pypdfium2 %s (libpdfium %s, %s)" % (
+            V_PYPDFIUM2, V_LIBPDFIUM,
+            "source build" if IS_SOURCEBUILD else "official build",
+        ),
     )
     
     subparsers = parser.add_subparsers(dest="subcommand")
     for cmd in Subcommands.values():
         cmd.attach_parser(subparsers)
     
-    if have_argcomplete:
+    if argcomplete is not None:
         argcomplete.autocomplete(parser)
     
     return parser.parse_args(argv)
 
 
 def main():
+    
+    lib_logger = logging.getLogger("pypdfium2")
+    lib_logger.addHandler(logging.StreamHandler())
+    
     args = parse_args()
     if not args.subcommand:
         print("One of the following subcommands must be given: %s" % [sc for sc in Subcommands.keys()])
         return
+    
     Subcommands[args.subcommand].main(args)
