@@ -2,7 +2,6 @@
 # SPDX-License-Identifier: Apache-2.0 OR BSD-3-Clause
 
 import os
-import ast
 from os.path import (
     join,
     abspath,
@@ -13,32 +12,11 @@ from pypdfium2 import _namespace as pdfium
 from pypdfium2._cli._parsers import pagetext_type
 
 
-def rotation_type(string):
-    rotation = int(string)
-    if rotation not in (0, 90, 180, 270):
-        raise ValueError("Invalid rotation value %s" % rotation)
-    return rotation
-
-
-def colour_type(string):
-    if string.lower() == "none":
-        return
-    else:
-        colour = ast.literal_eval(string)
-        if not isinstance(colour, (tuple, list)):
-            raise ValueError("Invalid colour type %s. Must be list or tuple." % type(colour))
-        if not len(colour) in (3, 4):
-            raise ValueError("Invalid number of colour values. Must be 3 or 4.")
-        if not all(isinstance(val, int) and 0 <= val <= 255 for val in colour):
-            raise ValueError("Colour values must be integers ranging from 0 to 255.")
-        return colour
-
-
-def crop_type(string):
-    crop = ast.literal_eval(string)
-    if not isinstance(crop, (tuple, list)) or len(crop) != 4 or not all(isinstance(c, (int, float)) for c in crop):
-        raise ValueError("Crop must be a list of four numbers.")
-    return crop
+ColourOpts = dict(
+    metavar = "C",
+    nargs = 4,
+    type = int,
+)
 
 
 def attach_parser(subparsers):
@@ -65,7 +43,7 @@ def attach_parser(subparsers):
     parser.add_argument(
         "--format", "-f",
         default = "jpg",
-        help = "File extension of the image format to use",
+        help = "The image format to use",
     )
     parser.add_argument(
         "--pages",
@@ -82,19 +60,30 @@ def attach_parser(subparsers):
     parser.add_argument(
         "--rotation",
         default = 0,
-        type = rotation_type,
+        type = int,
+        choices = (0, 90, 180, 270),
         help = "Rotate pages by 90, 180 or 270 degrees",
     )
     parser.add_argument(
-        "--colour",
+        "--fill-colour",
         default = (255, 255, 255, 255),
-        type = colour_type,
-        help = "Page background colour. Defaults to white. It can be given in RGBA format as a sequence of integers ranging from 0 to 255, or it may be 'none' for transparent background."
+        help = "Colour the bitmap will be filled with before rendering. It shall be given in RGBA format as a sequence of integers ranging from 0 to 255. Defaults to white.",
+        **ColourOpts,
+    )
+    parser.add_argument(
+        "--force-halftone",
+        action = "store_true",
+        help = "Always use halftone for image stretching",
     )
     parser.add_argument(
         "--no-annotations",
         action = "store_true",
-        help = "Option to prevent rendering of PDF annotations",
+        help = "Prevent rendering of PDF annotations",
+    )
+    parser.add_argument(
+        "--no-forms",
+        action = "store_true",
+        help = "Prevent rendering of PDF forms",
     )
     parser.add_argument(
         "--optimise-mode",
@@ -109,9 +98,10 @@ def attach_parser(subparsers):
     )
     parser.add_argument(
         "--crop",
-        type = crop_type,
+        nargs = 4,
+        type = float,
         default = (0, 0, 0, 0),
-        help = "Amount to crop from (left, bottom, right, top)"
+        help = "Amount to crop from (left, bottom, right, top)",
     )
     parser.add_argument(
         "--no-antialias",
@@ -120,11 +110,47 @@ def attach_parser(subparsers):
         choices = ("text", "image", "path"),
         help = "Item types that shall not be smoothed",
     )
+    parser.add_argument(
+        "--rev-byteorder",
+        action = "store_true",
+        help = "Render with reverse byte order internally, i. e. RGB(A) instead of BGR(A). The result should be completely identical.",
+    )
+    parser.add_argument(
+        "--prefer-bgrx",
+        action = "store_true",
+        help = "Request the use of a four-channel pixel format for coloured output, even if rendering without transparency.",
+    )
     parser.add_argument(
         "--processes",
         default = os.cpu_count(),
         type = int,
-        help = "The number of processes to use for rendering (defaults to the number of CPU cores)"
+        help = "The number of processes to use for rendering (defaults to the number of CPU cores)",
+    )
+    
+    colour_scheme = parser.add_argument_group(
+        title = "Colour scheme",
+        description = "Options for rendering with custom colour scheme",
+    )
+    colour_scheme.add_argument(
+        "--path-fill",
+        **ColourOpts
+    )
+    colour_scheme.add_argument(
+        "--path-stroke",
+        **ColourOpts
+    )
+    colour_scheme.add_argument(
+        "--text-fill",
+        **ColourOpts
+    )
+    colour_scheme.add_argument(
+        "--text-stroke",
+        **ColourOpts
+    )
+    colour_scheme.add_argument(
+        "--fill-to-stroke",
+        action = "store_true",
+        help = "Whether fill paths need to be stroked.",
     )
 
 
@@ -141,22 +167,44 @@ def main(args):
         else:
             page_indices = [i for i in range(len(pdf))]
         
-        n_digits = len(str( max(page_indices)+1 ))
+        cs_kwargs = dict(
+            path_fill = args.path_fill,
+            path_stroke = args.path_stroke,
+            text_fill = args.text_fill,
+            text_stroke = args.text_stroke,
+        )
+        cs = None
+        if all(cs_kwargs.values()):
+            cs = pdfium.ColourScheme(
+                fill_to_stroke = args.fill_to_stroke,
+                **cs_kwargs,
+            )
+        elif any(cs_kwargs.values()):
+            raise ValueError("If rendering with custom colour scheme, all parameters need to be set explicitly.")
         
-        renderer = pdf.render_topil(
+        kwargs = dict(
             page_indices = page_indices,
+            n_processes = args.processes,
             scale = args.scale,
             rotation = args.rotation,
             crop = args.crop,
-            colour = args.colour,
-            annotations = not args.no_annotations,
             greyscale = args.greyscale,
+            fill_colour = args.fill_colour,
+            colour_scheme = cs,
             optimise_mode = args.optimise_mode,
-            no_antialias = args.no_antialias,
-            n_processes = args.processes,
+            draw_annots = not args.no_annotations,
+            draw_forms = not args.no_forms,
+            force_halftone = args.force_halftone,
+            rev_byteorder = args.rev_byteorder,
+            prefer_bgrx = args.prefer_bgrx,
         )
+        for type in args.no_antialias:
+            kwargs["no_smooth%s" % type] = True
         
         prefix = splitext(basename(input_path))[0] + "_"
+        n_digits = len(str( max(page_indices)+1 ))
+        renderer = pdf.render_to(pdfium.BitmapConv.pil_image, **kwargs)
+        
         for image, index in zip(renderer, page_indices):
             suffix = str(index+1).zfill(n_digits)
             output_path = "%s.%s" % (join(args.output, prefix+suffix), args.format)
