@@ -35,6 +35,10 @@ def attach_parser(subparsers):
         "input",
         help = "PDF file on which to perform N-up compositing",
     )
+    parser.add_argument(
+        "--password",
+        help = "Password to unlock the PDF, if encrypted"
+    )
     parser.add_argument(
         "--output", "-o",
         required = True,
@@ -78,7 +82,7 @@ def main(args):
     width = units_to_pt(args.width, args.unit)
     height = units_to_pt(args.height, args.unit)
     
-    src_pdf = pdfium.PdfDocument(args.input)
+    src_pdf = pdfium.PdfDocument(args.input, password=args.password)
     raw_dest = pdfium.FPDF_ImportNPagesToOne(
         src_pdf.raw,
         width, height,
@@ -88,4 +92,4 @@ def main(args):
     with open(args.output, "wb") as buffer:
         dest_pdf.save(buffer)
     
-    [g.close() for g in (dest_pdf, src_pdf)]
+    for g in (dest_pdf, src_pdf): g.close()
