@@ -3,7 +3,11 @@
 
 import ctypes
 import pypdfium2._pypdfium as pdfium
-from pypdfium2._helpers.misc import raise_error
+from pypdfium2._helpers.misc import (
+    get_functype,
+    PdfiumError,
+    ErrorToStr,
+)
 
 
 class BufferDataHolder:
@@ -26,15 +30,15 @@ class ByteDataHolder:
         id(self.bytedata)
 
 
-class ReaderClass:
+class _reader_class:
     
     def __init__(self, buffer):
         self._buffer = buffer
     
     def __call__(self, _, position, p_buf, size):
-        c_buf = (ctypes.c_char * size).from_address( ctypes.addressof(p_buf.contents) )
+        c_buf = ctypes.cast(p_buf, ctypes.POINTER(ctypes.c_char * size))
         self._buffer.seek(position)
-        self._buffer.readinto(c_buf)
+        self._buffer.readinto(c_buf.contents)
         return 1
 
 
@@ -48,13 +52,12 @@ def open_pdf_buffer(buffer, password=None):
     file_len = buffer.tell()
     buffer.seek(0)
     
-    FuncType = ctypes.CFUNCTYPE(ctypes.c_int, ctypes.POINTER(None), ctypes.c_ulong, ctypes.POINTER(ctypes.c_ubyte), ctypes.c_ulong)
-    
     fileaccess = pdfium.FPDF_FILEACCESS()
     fileaccess.m_FileLen = file_len
-    fileaccess.m_GetBlock = FuncType( ReaderClass(buffer) )
+    fileaccess.m_GetBlock = get_functype(pdfium.FPDF_FILEACCESS, "m_GetBlock")( _reader_class(buffer) )
+    fileaccess.m_Param = None
     
-    pdf = pdfium.FPDF_LoadCustomDocument(ctypes.byref(fileaccess), password)
+    pdf = pdfium.FPDF_LoadCustomDocument(fileaccess, password)
     ld_data = BufferDataHolder(fileaccess.m_GetBlock, buffer)
     
     return pdf, ld_data
@@ -68,9 +71,12 @@ def open_pdf_bytes(bytedata, password=None):
 
 def open_pdf(input_data, password=None):
     
+    if isinstance(password, str):
+        password = password.encode("utf-8")
+    
     ld_data = None
     if isinstance(input_data, str):
-        pdf = pdfium.FPDF_LoadDocument(input_data, password)
+        pdf = pdfium.FPDF_LoadDocument(input_data.encode("utf-8"), password)
     elif isinstance(input_data, bytes):
         pdf, ld_data = open_pdf_bytes(input_data, password)
     elif is_input_buffer(input_data):
@@ -79,6 +85,8 @@ def open_pdf(input_data, password=None):
         raise TypeError("Invalid input type '%s'" % type(input_data).__name__)
     
     if pdfium.FPDF_GetPageCount(pdf) < 1:
-        raise_error("Loading the document failed")
+        err_code = pdfium.FPDF_GetLastError()
+        pdfium_msg = ErrorToStr.get(err_code, "Error code %s" % err_code)
+        raise PdfiumError("Loading the document failed (PDFium: %s)" % pdfium_msg)
     
     return pdf, ld_data
