@@ -10,29 +10,29 @@ import functools
 from concurrent.futures import ProcessPoolExecutor
 
 import pypdfium2._pypdfium as pdfium
-from pypdfium2._helpers._utils import ViewmodeMapping
 from pypdfium2._helpers._opener import (
     open_pdf,
     is_input_buffer,
 )
-from pypdfium2._helpers.page import PdfPage
 from pypdfium2._helpers.misc import (
     OutlineItem,
     FileAccess,
     PdfiumError,
+    ViewmodeToStr,
+    get_functype,
 )
+from pypdfium2._helpers.converters import BitmapConvAliases
+from pypdfium2._helpers.page import PdfPage
 
 try:
     import uharfbuzz as harfbuzz
 except ImportError:
-    have_harfbuzz = False
-else:
-    have_harfbuzz = True
+    harfbuzz = None
 
 logger = logging.getLogger(__name__)
 
 
-class PdfDocument:
+class PdfDocument (BitmapConvAliases):
     """
     Document helper class.
     
@@ -42,11 +42,16 @@ class PdfDocument:
             A byte buffer is defined as an object that implements the methods ``seek()``, ``tell()``, ``read()`` and ``readinto()``.
         password (str | bytes):
             A password to unlock the PDF, if encrypted.
+            If the document is not encrypted but a password was given, PDFium will ignore it.
         file_access (FileAccess):
             This parameter may be used to control how files are opened internally. It is ignored if *input_data* is not a file path.
         autoclose (bool):
             If set to :data:`True` and a byte buffer was provided as input, :meth:`.close` will not only close the PDFium document, but also the input source.
     
+    Raises:
+        PdfiumError: Raised if the document failed to load. The exception message is annotated with the reason reported by PDFium.
+        FileNotFoundError: Raised if an invalid or non-existent file path was given.
+    
     Hint:
         * :func:`len` may be called to get a document's number of pages.
         * :class:`PdfDocument` implements the context manager API, hence documents can be used in a ``with`` block, where :meth:`.close` will be called automatically on exit.
@@ -54,9 +59,8 @@ class PdfDocument:
         * Pages may be loaded using list index access.
         * The ``del`` keyword and list index access may be used to delete pages.
     
-    Note:
-        :class:`.PdfDocument` does not implement a page cache. This is to ensure correct behaviour in case the order or number of pages is modified using the raw API.
-        Therefore, it is up to the caller to cache pages appropriately and avoid inefficient repated loading/closing of pages.
+    Attributes:
+        raw (FPDF_DOCUMENT): The underlying PDFium document handle.
     """
     
     def __init__(
@@ -71,6 +75,8 @@ class PdfDocument:
         self._actual_input = input_data
         self._rendering_input = None
         self._ld_data = None
+        self._form_env = None
+        self._form_config = None
         
         self._password = password
         self._file_access = file_access
@@ -92,12 +98,12 @@ class PdfDocument:
                 self._actual_input = buf.read()
                 buf.close()
             else:
-                raise ValueError("An invalid file access strategy was given: '%s'" % self._file_access)
+                assert False
         
         if isinstance(self._actual_input, pdfium.FPDF_DOCUMENT):
-            self._pdf = self._actual_input
+            self.raw = self._actual_input
         else:
-            self._pdf, self._ld_data = open_pdf(self._actual_input, self._password)
+            self.raw, self._ld_data = open_pdf(self._actual_input, self._password)
     
     
     def __enter__(self):
@@ -107,7 +113,7 @@ class PdfDocument:
         self.close()
     
     def __len__(self):
-        return pdfium.FPDF_GetPageCount(self._pdf)
+        return pdfium.FPDF_GetPageCount(self.raw)
     
     def __iter__(self):
         for i in range( len(self) ):
@@ -120,11 +126,6 @@ class PdfDocument:
         self.del_page(i)
     
     
-    @property
-    def raw(self):
-        """ FPDF_DOCUMENT: The raw PDFium document object handle. """
-        return self._pdf
-    
     @classmethod
     def new(cls):
         """
@@ -139,14 +140,64 @@ class PdfDocument:
         Close the document to release allocated memory.
         This function shall be called when finished working with the object.
         """
-        pdfium.FPDF_CloseDocument(self._pdf)
+        
+        self.exit_formenv()
+        
+        pdfium.FPDF_CloseDocument(self.raw)
+        self.raw = None
+        
         if self._ld_data is not None:
             self._ld_data.close()
         if self._autoclose and is_input_buffer(self._actual_input):
             self._actual_input.close()
     
     
-    def save(self, buffer, version=None, rm_security=False):
+    def init_formenv(self):
+        """
+        Initialise a form environment handle for this document.
+        If already initialised, the existing one will be returned instead.
+        
+        Returns:
+            FPDF_FORMHANDLE:
+        """
+        if self._form_env is not None:
+            return self._form_env
+        self._form_config = pdfium.FPDF_FORMFILLINFO()
+        self._form_config.version = 2
+        self._form_env = pdfium.FPDFDOC_InitFormFillEnvironment(self.raw, self._form_config)
+        return self._form_env
+    
+    
+    def exit_formenv(self):
+        """
+        Release allocated memory by exiting the form environment.
+        If the form environment is not initialised, nothing will be done.
+        
+        Note:
+            This method is called by :meth:`.close`.
+        """
+        if self._form_env is None:
+            return
+        pdfium.FPDFDOC_ExitFormFillEnvironment(self._form_env)
+        id(self._form_config)
+        self._form_env = None
+        self._form_config = None
+    
+    
+    def get_version(self):
+        """
+        Returns:
+            int | None: The PDF version of the document (14 for 1.4, 15 for 1.5, ...),
+            or :data:`None` if the version could not be determined (e. g. because the document was created using :meth:`PdfDocument.new`).
+        """
+        version = ctypes.c_int()
+        success = pdfium.FPDF_GetFileVersion(self.raw, version)
+        if not success:
+            return
+        return int(version.value)
+    
+    
+    def save(self, buffer, version=None):
         """
         Save the document into an output buffer, at its current state.
         
@@ -154,24 +205,16 @@ class PdfDocument:
             buffer (typing.BinaryIO):
                 A byte buffer to capture the data.
                 It may be any object implementing the ``write()`` method.
-            version (typing.Optional[int]):
+            version (int | None):
                  The PDF version to use, given as an integer (14 for 1.4, 15 for 1.5, ...).
                  If :data:`None`, PDFium will set a version automatically.
-            rm_security (bool):
-                If this option is :data:`True` and an encrypted PDF is saved, its password protection will be removed.
-                Otherwise, encryption will be retained.
         """
         
-        WriteFunctype = ctypes.CFUNCTYPE(ctypes.c_int, ctypes.POINTER(pdfium.FPDF_FILEWRITE), ctypes.POINTER(None), ctypes.c_ulong)
-        
         filewrite = pdfium.FPDF_FILEWRITE()
-        filewrite.WriteBlock = WriteFunctype( _writer_class(buffer) )
+        filewrite.version = 1
+        filewrite.WriteBlock = get_functype(pdfium.FPDF_FILEWRITE, "WriteBlock")( _writer_class(buffer) )
         
-        flags = pdfium.FPDF_NO_INCREMENTAL
-        if rm_security:
-            flags |= pdfium.FPDF_REMOVE_SECURITY
-        
-        saveargs = (self._pdf, ctypes.byref(filewrite), flags)
+        saveargs = (self.raw, filewrite, pdfium.FPDF_NO_INCREMENTAL)
         if version is None:
             success = pdfium.FPDF_SaveAsCopy(*saveargs)
         else:
@@ -181,10 +224,30 @@ class PdfDocument:
             raise PdfiumError("Saving the document failed")
     
     
-    def _verify_index(self, index):
+    def _handle_index(self, index):
         n_pages = len(self)
+        if index < 0:
+            index += n_pages
         if not 0 <= index < n_pages:
             raise IndexError("Page index %s is out of bounds for document with %s pages." % (index, n_pages))
+        return index
+    
+    
+    def get_page_size(self, index):
+        """
+        Get the dimensions of the page at *index*. Reverse indexing is allowed.
+        This method is equivalent to :meth:`.PdfPage.get_size()` but works without needing to load a separate page object.
+        
+        Returns:
+            (float, float): Page width and height in PDF canvas units.
+        """
+        index = self._handle_index(index)
+        size = pdfium.FS_SIZEF()
+        success = pdfium.FPDF_GetPageSizeByIndexF(self.raw, index, size)
+        if not success:
+            raise PdfiumError("Getting page size by index failed.")
+        return (float(size.width), float(size.height))
+    
     
     def new_page(self, width, height, index=None):
         """
@@ -195,31 +258,37 @@ class PdfDocument:
                 Target page width (horizontal size).
             height (float):
                 Target page height (vertical size).
-            index (typing.Optional[int]):
+            index (int | None):
                 Suggested zero-based index at which the page will be inserted.
-                If *index* is less or equal to zero, the page will be inserted at the beginning.
+                If *index* is negative, the indexing direction will be reversed.
+                If *index* is zero, the page will be inserted at the beginning.
                 If *index* is :data:`None` or larger that the document's current last index, the page will be appended to the end.
-        
         Returns:
             PdfPage: The newly created page.
         """
         if index is None:
             index = len(self)
-        raw_page = pdfium.FPDFPage_New(self._pdf, index, width, height)
+        elif index < 0:
+            index += len(self)
+        raw_page = pdfium.FPDFPage_New(self.raw, index, width, height)
         return PdfPage(raw_page, self)
     
+    
     def del_page(self, index):
-        """ Remove the page at *index*. """
-        self._verify_index(index)
-        pdfium.FPDFPage_Delete(self._pdf, index)
+        """
+        Remove the page at *index*. Reverse indexing is allowed.
+        """
+        index = self._handle_index(index)
+        pdfium.FPDFPage_Delete(self.raw, index)
+    
     
     def get_page(self, index):
         """
         Returns:
-            PdfPage: The page at *index*.
+            PdfPage: The page at *index*. Reverse indexing is allowed.
         """
-        self._verify_index(index)
-        raw_page = pdfium.FPDF_LoadPage(self._pdf, index)
+        index = self._handle_index(index)
+        raw_page = pdfium.FPDF_LoadPage(self.raw, index)
         return PdfPage(raw_page, self)
     
     
@@ -231,7 +300,7 @@ class PdfDocument:
             font_path (str):
                 File path of the font to use.
             type (int):
-                A constant signifying the type of the given font (:data:`.FPDF_FONT_TYPE1` or :data:`.FPDF_FONT_TRUETYPE`).
+                A constant signifying the type of the given font (:data:`.FPDF_FONT_*`).
             is_cid (bool):
                 Whether the given font is a CID font or not.
         Returns:
@@ -242,26 +311,36 @@ class PdfDocument:
             font_data = fh.read()
         
         pdf_font = pdfium.FPDFText_LoadFont(
-            self._pdf,
+            self.raw,
             ctypes.cast(font_data, ctypes.POINTER(ctypes.c_uint8)),
             len(font_data),
             type,
             is_cid,
         )
         
-        return PdfFont(pdf_font, font_data)
+        return PdfFont(pdf_font, self, font_data)
     
     
     def _get_bookmark(self, bookmark, level):
         
-        t_buflen = pdfium.FPDFBookmark_GetTitle(bookmark, None, 0)
-        t_buffer = ctypes.create_string_buffer(t_buflen)
-        pdfium.FPDFBookmark_GetTitle(bookmark, t_buffer, t_buflen)
-        title = t_buffer.raw[:t_buflen].decode('utf-16-le')[:-1]
+        n_bytes = pdfium.FPDFBookmark_GetTitle(bookmark, None, 0)
+        buffer = ctypes.create_string_buffer(n_bytes)
+        pdfium.FPDFBookmark_GetTitle(bookmark, buffer, n_bytes)
+        title = buffer.raw[:n_bytes-2].decode('utf-16-le')
         
-        is_closed = pdfium.FPDFBookmark_GetCount(bookmark) < 0
-        dest = pdfium.FPDFBookmark_GetDest(self._pdf, bookmark)
-        page_index = pdfium.FPDFDest_GetDestPageIndex(self._pdf, dest)
+        count = pdfium.FPDFBookmark_GetCount(bookmark)
+        if count < 0:
+            is_closed = True
+        elif count == 0:
+            is_closed = None
+        else:
+            is_closed = False
+        
+        n_kids = abs(count)
+        dest = pdfium.FPDFBookmark_GetDest(self.raw, bookmark)
+        page_index = pdfium.FPDFDest_GetDestPageIndex(self.raw, dest)
+        if page_index == -1:
+            page_index = None
         
         n_params = ctypes.c_ulong()
         view_pos = (pdfium.FS_FLOAT * 4)()
@@ -272,6 +351,7 @@ class PdfDocument:
             level = level,
             title = title,
             is_closed = is_closed,
+            n_kids = n_kids,
             page_index = page_index,
             view_mode = view_mode,
             view_pos = view_pos,
@@ -295,12 +375,10 @@ class PdfDocument:
             :class:`.OutlineItem`: The data of an outline item ("bookmark").
         """
         
-        if level >= max_depth:
-            return []
         if seen is None:
             seen = set()
         
-        bookmark = pdfium.FPDFBookmark_GetFirstChild(self._pdf, parent)
+        bookmark = pdfium.FPDFBookmark_GetFirstChild(self.raw, parent)
         
         while bookmark:
             
@@ -312,14 +390,15 @@ class PdfDocument:
                 seen.add(address)
             
             yield self._get_bookmark(bookmark, level)
-            yield from self.get_toc(
-                max_depth = max_depth,
-                parent = bookmark,
-                level = level + 1,
-                seen = seen,
-            )
+            if level < max_depth-1:
+                yield from self.get_toc(
+                    max_depth = max_depth,
+                    parent = bookmark,
+                    level = level + 1,
+                    seen = seen,
+                )
             
-            bookmark = pdfium.FPDFBookmark_GetNextSibling(self._pdf, bookmark)
+            bookmark = pdfium.FPDFBookmark_GetNextSibling(self.raw, bookmark)
     
     
     @staticmethod
@@ -335,13 +414,24 @@ class PdfDocument:
         """
         
         for item in toc:
+            
+            if item.n_kids == 0:
+                state = "*"
+            elif item.is_closed:
+                state = "-"
+            else:
+                state = "+"
+            
+            if item.page_index is None:
+                target = "?"
+            else:
+                target = item.page_index + 1
+            
             print(
                 "    " * item.level +
-                "[%s] " % ("-" if item.is_closed else "+") +
-                "%s -> %s  # %s %s" % (
-                    item.title,
-                    item.page_index+1,
-                    ViewmodeMapping[item.view_mode],
+                "[%s] %s -> %s  # %s %s" % (
+                    state, item.title, target,
+                    ViewmodeToStr[item.view_mode],
                     [round(c, n_digits) for c in item.view_pos],
                 )
             )
@@ -351,7 +441,7 @@ class PdfDocument:
         """
         Update the input sources for concurrent rendering to the document's current state
         by saving to bytes and setting the result as new input.
-        If you modified the document, you may want to call this method before :meth:`._render_base`.
+        If you modified the document, you may want to call this method before :meth:`.render_to`.
         """
         buffer = io.BytesIO()
         self.save(buffer)
@@ -361,39 +451,38 @@ class PdfDocument:
     
     
     @classmethod
-    def _process_page(cls, index, renderer_name, input_data, password, file_access, **kwargs):
+    def _process_page(cls, index, converter, input_data, password, file_access, **kwargs):
         pdf = cls(
             input_data,
             password = password,
             file_access = file_access,
         )
         page = pdf.get_page(index)
-        result = getattr(page, "render_to"+renderer_name)(**kwargs)
+        result = page.render_to(converter, **kwargs)
         for g in (page, pdf): g.close()
         return result, index
     
     
-    def _render_base(
+    def render_to(
             self,
-            renderer_name,
+            converter,
             page_indices = None,
             n_processes = os.cpu_count(),
             **kwargs
         ):
         """
         Concurrently render multiple pages, using a process pool executor.
-        This method serves as base for :meth:`.render_tobytes` and :meth:`render_topil`. Embedders should never call it directly.
         
-        The order of results matches the order of given page indices.
+        .. seealso:: :meth:`.PdfPage.render_to` / :meth:`.PdfPage.render_base`
         
         Parameters:
-            page_indices (typing.Optional[typing.Sequence[int]]):
-                A sequence of zero-based indices of the pages to render.
-                If :data:`None`, all pages will be included.
+            page_indices (typing.Sequence[int] | None):
+                A sequence of zero-based indices of the pages to render. Reverse indexing or duplicate page indices are prohibited.
+                If :data:`None`, all pages will be included. The order of results is guaranteed to match the order of given page indices.
             n_processes (int):
                 Target number of parallel processes.
             kwargs (dict):
-                Keyword arguments to be passed to :meth:`.PdfPage.render_base`.
+                Keyword arguments to the renderer.
         Yields:
             :data:`typing.Any`: Implementation-specific result object.
         """
@@ -412,15 +501,17 @@ class PdfDocument:
                 self._rendering_input = self._orig_input
         
         n_pages = len(self)
-        
-        if page_indices is None or len(page_indices) == 0:
+        if not page_indices:
             page_indices = [i for i in range(n_pages)]
-        if not all(0 <= i < n_pages for i in page_indices):
-            raise ValueError("A page index is out of bounds.")
+        else:
+            if not all(0 <= i < n_pages for i in page_indices):
+                raise ValueError("Out-of-bounds page index")
+            if len(page_indices) != len(set(page_indices)):
+                raise ValueError("Duplicate page index")
         
         invoke_renderer = functools.partial(
             PdfDocument._process_page,
-            renderer_name = renderer_name,
+            converter = converter,
             input_data = self._rendering_input,
             password = self._password,
             file_access = self._file_access,
@@ -435,29 +526,6 @@ class PdfDocument:
                 yield result
         
         assert len(page_indices) == i
-    
-    
-    def render_tobytes(self, **kwargs):
-        """
-        Concurrently render pages to bytes.
-        See :meth:`.PdfDocument._render_base` and :meth:`.PdfPage.render_base` for possible keyword arguments.
-        
-        Yields:
-            :class:`tuple`: Result of :meth:`.PdfPage.render_tobytes`.
-        """
-        yield from self._render_base("bytes", **kwargs)
-    
-    def render_topil(self, **kwargs):
-        """
-        *Requires* :mod:`PIL`.
-        
-        Concurrently render pages to PIL images.
-        See :meth:`.PdfDocument._render_base` and :meth:`.PdfPage.render_base` for possible keyword arguments.
-        
-        Yields:
-            :class:`PIL.Image.Image`: PIL image.
-        """
-        yield from self._render_base("pil", **kwargs)
 
 
 class _writer_class:
@@ -477,7 +545,7 @@ class HarfbuzzFont:
     """ Harfbuzz font data helper class. """
     
     def __init__(self, font_path):
-        if not have_harfbuzz:
+        if harfbuzz is None:
             raise RuntimeError("Font helpers require uharfbuzz to be installed.")
         self.blob = harfbuzz.Blob.from_file_path(font_path)
         self.face = harfbuzz.Face(self.blob)
@@ -486,21 +554,24 @@ class HarfbuzzFont:
 
 
 class PdfFont:
-    """ PDF font data helper class. """
+    """
+    PDF font data helper class.
     
-    def __init__(self, pdf_font, font_data):
-        self._pdf_font = pdf_font
+    Attributes:
+        raw (FPDF_FONT): The underlying PDFium font handle.
+        pdf (PdfDocument): Reference to the document this font belongs to.
+    """
+    
+    def __init__(self, raw, pdf, font_data):
+        self.raw = raw
+        self.pdf = pdf
         self._font_data = font_data
     
-    @property
-    def raw(self):
-        """ FPDF_FONT: The raw PDFium font object handle. """
-        return self._pdf_font
-    
     def close(self):
         """
         Close the font to release allocated memory.
         This function shall be called when finished working with the object.
         """
-        pdfium.FPDFFont_Close(self._pdf_font)
+        pdfium.FPDFFont_Close(self.raw)
+        self.raw = None
         id(self._font_data)
